import logging
import math
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import date, timedelta
from typing import List

from requests.exceptions import HTTPError

from .repo import (get_all_pkg_names, get_pkg_github_repo, get_pkg_pypi_entry,
                   get_pkg_stats)
from .types import Package, ProgressBar
from .types.github_repo import GithubRepo
from .types.pepy_tech import PackageStats
from .types.pypi_entry import PypiEntry


def find_packages(name_search: str, progressbar: ProgressBar, threads: int = 10) -> List[Package]:
    search_phrases = name_search.lower().split(',')
    def name_matches_phrases(pkg_name: str) -> bool:
        return all(
            search_phrase in pkg_name
            for search_phrase in search_phrases
        )

    all_pkg_names = [name.lower() for name in get_all_pkg_names()]
    matching_pkg_names = [name for name in all_pkg_names if name_matches_phrases(name)]

    MAX_MATCHING_NAMES = 1000
    assert len(matching_pkg_names) <= MAX_MATCHING_NAMES, f'Too many results to consider (max={MAX_MATCHING_NAMES}), try to use more precise filter'

    with ThreadPoolExecutor(threads) as executor:
        futures = [
            executor.submit(get_package_info, pkg_name)
            for pkg_name in matching_pkg_names
        ]

        with progressbar(as_completed(futures), len(futures)) as bar:
            return [
                future.result()
                for future in bar
            ]


def get_package_info(name: str) -> Package:
    pkg = Package(name=name)

    try:
        pypi_info: PypiEntry = get_pkg_pypi_entry(name)
    except HTTPError as e:
        logging.warn(f'failed to get pypi info for {name}: {e}', RuntimeWarning)
    else:
        info = pypi_info.info
        pkg.summary = info.summary
        pkg.version = info.version
        pkg.home_page = info.home_page

        if info.project_urls:
            src_url = info.project_urls.Source
            if src_url and any(
                    vcs_name in src_url
                    for vcs_name in ['github', 'bitbucket', 'gitlab']
                ):
                pkg.home_page = src_url

        releases = pypi_info.releases
        if num_releases := len(releases):
            pkg.releases = num_releases
            if upload_days := {
                upload.upload_time.date()
                for uploads in releases.values()
                for upload in uploads
            }:
                pkg.last_release_date = str(max(upload_days))

    if pkg.releases:
        try:
            stats: PackageStats = get_pkg_stats(name)
        except HTTPError as e:
            logging.warn(f'failed to get downloads info for {name}: {e}', RuntimeWarning)
        else:
            all_downloads = stats.downloads
            day_from = date.today() - timedelta(days=90)
            recent_downloads: int = sum([
                sum(version_downloads.values(), 0)
                for stats_day, version_downloads in all_downloads.items()
                if stats_day > day_from
            ], 0)

            pkg.downloads = recent_downloads


    if pkg.home_page and ('github.com' in pkg.home_page):
        try:
            github_repo: GithubRepo = get_pkg_github_repo(pkg.home_page)
        except HTTPError as e:
            logging.warn(f'failed to get github info for {name}: {e}', RuntimeWarning)
        else:
            pkg.stars = github_repo.stargazers_count

    pkg.score = _get_score(pkg)

    return pkg


def _get_score(package: Package) -> int:
    if not package.last_release_date:
        return 0

    score: float = 0

    if package.downloads:
        score += math.log10(package.downloads)

    if package.stars:
        score +=  math.log2(package.stars)

    days_from_last_release = (date.today() - date.fromisoformat(package.last_release_date)).days
    if days_from_last_release <= 30:
        score += 3
    elif days_from_last_release <= 60:
        score += 2
    elif days_from_last_release <= 120:
        score += 1
    elif days_from_last_release <= 180:
        pass
    elif days_from_last_release <= 270:
        score -= 1
    elif days_from_last_release <= 360:
        score -= 2
    else:
        score -= 3

    return round(score)



import json
import logging
from importlib.metadata import version
from operator import attrgetter
from typing import Any, List, Optional, TypeVar

import click
from tabulate import tabulate

from . import cache, user_config
from .github_auth import AccessToken, github_device_flow
from .service import find_packages
from .types import Package


@click.group()
@click.version_option(version('pypi-client'))
def cli() -> None:
    pass


@cli.command()
def auth_github() -> None:
    """Log into GitHub"""

    def write_token(access_token: AccessToken) -> None:
        user_config.write_oauth_token(access_token.access_token)

    with github_device_flow(write_token) as verif_codes:
        print(f'Please open {verif_codes.verification_uri} and enter code: {verif_codes.user_code}')
    print('Success')


def validate_pkg_name(ctx: Any, param: Any, value: str) -> str:
    MIN_LEN = 3
    if len(value) < MIN_LEN:
        raise click.BadParameter(f'name too short, min length={MIN_LEN}')

    return value


@cli.command()
@click.argument('name-search', callback=validate_pkg_name)
@click.option('--limit', type=click.IntRange(min=1), help='Max number of items to return')
@click.option('--no-cache', is_flag=True, type=click.BOOL, default=False, help='Clear cache before run')
@click.option('--verbose', is_flag=True, type=click.BOOL, default=False, help='Print debug messages')
@click.option('--json', "as_json", is_flag=True, type=click.BOOL, default=False, help='Return in json format')
@click.option('--threads', type=click.INT, default=10, help='Number of threads to use')
def search(name_search: str, limit: int, no_cache: bool, verbose: bool, as_json: bool, threads: int) -> None:
    """Search python package by name"""

    if no_cache:
        cache.clear()

    logging.basicConfig(level=logging.DEBUG if verbose else logging.ERROR)

    try:
        found_packages = find_packages(name_search, click.progressbar, threads)
    except AssertionError as e:
        print(e)
        return

    sorted_packages = list(reversed(sorted(found_packages, key=attrgetter('score'))))
    if limit:
        sorted_packages = sorted_packages[:limit]

    print_func = _print_as_json if as_json else _print_as_text
    print_func(sorted_packages)


def _print_as_json(sorted_packages: List[Package]) -> None:
    for pkg in sorted_packages:
        print(json.dumps(pkg.__dict__))


def _print_as_text(sorted_packages: List[Package]) -> None:
    if pkg_count := len(sorted_packages):
        print(f'Found {pkg_count} packages:')
    else:
        print('No matching packages found')
        return

    columns = ['name', 'downloads', 'summary', 'version', 'home_page', 'stars', 'releases', 'last_release_date']
    columns_max_width = {
        'name': 30,
        'summary': 50,
        'home_page': 50
    }

    Val = TypeVar('Val', str, int)
    def _enforse_max_width(val: Val, max_width: Optional[int]) -> Val:
        if max_width and isinstance(val, str) and len(val) > max_width:
            return val[:max_width] + '...'

        return val


    print(tabulate([
        [
            _enforse_max_width(getattr(pkg, col), columns_max_width.get(col))
            for col in columns
        ]
        for pkg in sorted_packages
    ], headers=columns))


if __name__ == '__main__':
    cli()



import functools
import logging
from typing import Iterator

import requests
from lxml import html

from . import cache
from .types.github_repo import GithubRepo
from .types.pepy_tech import PackageStats
from .types.pypi_entry import PypiEntry
from .user_config import read_oauth_token


def get_all_pkg_names() -> Iterator[str]:   
    tree: html.HtmlElement = html.fromstring(_get_all_pkgs_html())
    return tree.xpath('//a/text()')


@cache.memoize()
def _get_all_pkgs_html() -> str:
    logging.debug('get_all_pkgs_html')
    response = requests.get("https://pypi.org/simple/")
    response.raise_for_status()
    return response.text 


@cache.memoize()
def get_pkg_pypi_entry(pkg_name: str) -> PypiEntry:
    logging.debug(f'get_pkg_pypi_entry for {pkg_name}')
    response = requests.get(f'https://pypi.org/pypi/{pkg_name}/json')
    response.raise_for_status()
    return PypiEntry(**response.json())    


@cache.memoize()
def get_pkg_stats(pkg_name: str, last_days: int = 90) -> PackageStats:
    logging.debug(f'get_pkg_stats for {pkg_name}')
    url = 'https://api.pepy.tech/api/v2/projects/' + pkg_name
    response = requests.get(url)
    response.raise_for_status()
    return PackageStats(**response.json())


@cache.memoize()
def get_pkg_github_repo(pkg_repo: str) -> GithubRepo:
    logging.debug(f'get_pkg_github_repo for {pkg_repo}')
    repo_name = pkg_repo.strip('/').split('/')[-2:]
    url = 'https://api.github.com/repos/' + '/'.join(repo_name)
    response = requests.get(url, auth=('token', _get_github_oauth_token()))
    response.raise_for_status()
    return GithubRepo(**response.json())


@functools.lru_cache(maxsize=None)
def _get_github_oauth_token() -> str:
    return read_oauth_token()



from dataclasses import dataclass
from typing import Callable, ContextManager, Iterable, Optional, TypeVar

from pydantic import BaseModel  # pylint: disable=no-name-in-module


class Package(BaseModel):
    name: str
    summary: Optional[str]
    version: Optional[str]
    home_page: Optional[str]
    downloads: Optional[int]
    stars: Optional[int]
    releases: Optional[int]
    last_release_date: Optional[str]
    score: Optional[int]


# Progress bar
Item = TypeVar('Item')
ProgressBar = Callable[[Iterable[Item], int], ContextManager[Iterable[Item]]]


class DeviceFlowVerificationCodes(BaseModel):
    device_code: str
    user_code: str
    verification_uri: str
    expires_in: int
    interval: int


class AccessToken(BaseModel):
    access_token: str



from datetime import date
from typing import Dict

from pydantic import BaseModel  # pylint: disable=no-name-in-module

Version = str
class PackageStats(BaseModel):
    downloads: Dict[date, Dict[Version, int]]



import logging
import time
from contextlib import contextmanager
from typing import Callable, Generator
from urllib import parse

import requests

from .types import AccessToken, DeviceFlowVerificationCodes

CLIENT_ID = 'da5e9528b63f1bd10fd8'


@contextmanager
def github_device_flow(write_oauth_token: Callable[[AccessToken], None]) -> Generator[DeviceFlowVerificationCodes, None, None]:
    verif_codes: DeviceFlowVerificationCodes = _get_verification_codes()
    yield verif_codes
    access_token: AccessToken = _wait_for_authorization(verif_codes.device_code, verif_codes.expires_in, verif_codes.interval)
    write_oauth_token(access_token)


def _get_verification_codes() -> DeviceFlowVerificationCodes:
    res = requests.post('https://github.com/login/device/code', {'client_id': CLIENT_ID})
    res.raise_for_status()
    r = dict(parse.parse_qsl(res.text))
    return DeviceFlowVerificationCodes(**r)


def _wait_for_authorization(device_code: str, expires_in: int, sleep_interval: int) -> AccessToken:
    time.sleep(sleep_interval)
    expires_in -= sleep_interval

    while expires_in > 0:
        try:
            return _get_access_token(device_code)
        except Exception as e:
            logging.debug(e)
            time.sleep(sleep_interval)
            expires_in -= sleep_interval
    else:
        raise Exception('Verification code expired')


def _get_access_token(device_code: str) -> AccessToken:
    res = requests.post('https://github.com/login/oauth/access_token', {
        'client_id': CLIENT_ID,
        'device_code': device_code,
        'grant_type': 'urn:ietf:params:oauth:grant-type:device_code'
    })
    res.raise_for_status()
    r = dict(parse.parse_qsl(res.text))

    if 'error' in r:
        raise Exception(r['error_description'], r['error'])

    return AccessToken(**r)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['pypi_client', 'pypi_client.types']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'diskcache>=5.0.3,<6.0.0',
 'lxml>=4.6.1,<5.0.0',
 'pydantic>=1.7.2,<2.0.0',
 'requests-mock>=1.8.0,<2.0.0',
 'requests>=2.24.0,<3.0.0',
 'tabulate>=0.8.7,<0.9.0']

entry_points = \
{'console_scripts': ['pypi-client = pypi_client.cli:cli']}

setup_kwargs = {
    'name': 'pypi-client',
    'version': '0.1.11',
    'description': 'PyPI command-line tool',
    'long_description': "PyPI client\n===========\n\n![Lint and test](https://github.com/abahdanovich/pypi-client/workflows/Lint%20and%20test/badge.svg)\n\nCLI tool for searching for a python package by name.\n\n* fetches all package names from PyPi\n* filters and finds matching packages (by name)\n* downloads github stars (if package uses GH as a repo) number and package downloads\n* shows results in a table or json\n\nInstall\n-------\n\n```\npip install pypi-client\n```\n\nUsage\n-----\n\n```\nUsage: pypi-client [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --version  Show the version and exit.\n  --help     Show this message and exit.\n\nCommands:\n  auth-github  Log into GitHub\n  search       Search python package by name\n```\n\nSearch\n------\n\n```\nUsage: pypi-client search [OPTIONS] NAME_SEARCH\n\n  Search python package by name\n\nOptions:\n  --limit INTEGER RANGE  Max number of items to return\n  --no-cache             Clear cache before run\n  --verbose              Print debug messages\n  --json                 Return in json format\n  --threads INTEGER      Number of threads to use\n  --help                 Show this message and exit.\n```\n\nExample output:\n\n```\n$ pypi-client search kafka\nFound 155 packages:\nname                                 downloads  summary                                                version      home_page                                                stars    releases  last_release_date\n---------------------------------  -----------  -----------------------------------------------------  -----------  -----------------------------------------------------  -------  ----------  -------------------\nkafka-python                           6863094  Pure Python client for Apache Kafka                    2.0.2        https://github.com/dpkp/kafka-python                      4084          34  2020-09-30\nconfluent-kafka                        3341286  Confluent's Python client for Apache Kafka             1.5.0        https://github.com/confluentinc/confluent-kafka-py...     2017          20  2020-08-07\nns-kafka-python                           5739  Pure Python client for Apache Kafka                    1.4.7        https://github.com/dpkp/kafka-python                      4084           1  2020-09-28\ntencentcloud-sdk-python-ckafka           11820  Tencent Cloud Ckafka SDK for Python                    3.0.290      https://github.com/TencentCloud/tencentcloud-sdk-p...      297          40  2020-11-12\nkafka                                   939197  Pure Python client for Apache Kafka                    1.3.5        https://github.com/dpkp/kafka-python                      4084          17  2017-10-07\n[...]\n```",
    'author': 'Andrzej Bogdanowicz',
    'author_email': 'bahdanovich@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/abahdanovich/pypi-client',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



from pathlib import Path

from appdirs import user_config_dir

config_dir = Path(user_config_dir('pypi-client', 'PyPI'))
oauth_token_file = 'github_oauth_token'


def write_oauth_token(oauth_token: str) -> None:
    config_dir.mkdir(parents=True, exist_ok=True)
    with open(config_dir / oauth_token_file, 'w') as f:
        f.write(oauth_token)


def read_oauth_token() -> str:
    try:
        with open(config_dir / oauth_token_file, 'r') as f:
            return f.readline()
    except:
        raise Exception('Github OAuth token not found, please use `auth-github` command to obtain token')



from datetime import datetime
from typing import Dict, List, Optional

from pydantic import BaseModel  # pylint: disable=no-name-in-module

Url = str
class PypiProjectUrls(BaseModel):
    Source: Optional[Url]

class PypiPackageInfo(BaseModel):
    summary: Optional[str]
    version: Optional[str] 
    project_urls: Optional[PypiProjectUrls]
    home_page: Optional[Url]

    # @validator('project_urls', pre=True)
    # def ensure_project_urls(cls, val: Any) -> Any:      # pylint: disable=no-self-argument
    #     return val or []

class PypiReleaseUpload(BaseModel):
    upload_time: datetime

Version = str
class PypiEntry(BaseModel):
    info: PypiPackageInfo
    releases: Dict[Version, List[PypiReleaseUpload]] = {}
